#!/usr/bin/env python3

from .errors import *

class Forum:

    def __init__(self):
        self.world = None
        self.actors = None

    def execute_message(self, message):
        info("executing a message: {message}")

        # Relay the messages to clients running on other machines, if this is a 
        # multiplayer game.  Since the tokens referenced in the message might 
        # be changed once the message is executed, the message has to be 
        # relayed before then.

        for actor in self.actors:
            actor._relay_message(message)

        # Normally, tokens can only call methods that have been decorated with 
        # @read_only.  This is a precaution to help keep the worlds in sync on 
        # all the clients.  This restriction is lifted when the tokens are 
        # handling messages and enforced again once the actors are handling 
        # messages.

        with self.world._unlock_temporarily():

            # First, let the message update the state of the game world.

            message._execute(self.world)

            # Second, let the world react to the message.  The main effect of 
            # the message should have already been carried out above.  These 
            # callbacks should take care of more peripheral effects.

            self.world._react_to_message(message)

        # Third, let the actors and the extensions react to the message.  This 
        # step is carried out last so that the actors can be sure that the 
        # world has a consistent state by the time their handlers are called.

        for actor in self.actors:
            actor._react_to_message(message)

    def connect_everyone(self, world, actors):
        # Save references to the world and the actors in the forum.

        self.world = world
        self.actors = actors

        # Save references to the actors in the world.  The world doesn't need 
        # to know about the forum because it can't send messages.  It needs to 
        # know about the actors so it can create token extensions.

        self.world._set_actors(actors)

        # Save references to the forum and the world in the actors.  Also 
        # assign each actor a factory it can use to generate unique token ids.
        #
        # In multiplayer games, each client needs the ability to create tokens, 
        # so that messages can be instantly handled.  Tokens still need unique 
        # ids though, so this method provides each actor with an IdFactory that 
        # generates ids using an offset and a spacing to ensure uniqueness.
        #
        # Actors take their own id numbers (used for figuring out who messages 
        # were sent by) from the offset parameter of the id factory.  Since the 
        # Referee must have an id of 0 if it's present, care is taken to make 
        # that happen.

        id_factories = self._assign_id_factories()

        for actor in self.actors:
            actor._set_world(world)
            actor._set_forum(self, id_factories[actor])

    def on_start_game(self):
        pass

    def on_update_game(self):
        # Forum doesn't do anything on a timer; it does everything in response 
        # to a message being sent.  But the ClientForum uses this method to 
        # react to message that have arrived from the server.
        pass

    def on_finish_game(self):
        pass

    def _assign_id_factories(self):
        id_factories = {}
        actors = sorted(self.actors, key=lambda x: not x.is_referee())
        first_id = self.world.get_last_id() + 1
        spacing = len(self.actors)

        for offset, actor in enumerate(actors, first_id):
            id_factories[actor] = IdFactory(offset, spacing)

        return id_factories


class ForumObserver:

    from collections import namedtuple
    CallbackInfo = namedtuple('CallbackInfo', 'message_cls, callback')

    def __init__(self):
        super().__init__()

        # Create a data structure to hold all the callbacks registered with 
        # this observer.  Using a dictionary to distinguish between the regular 
        # message handlers, the soft sync error handlers, and the hard sync 
        # error handlers (instead of just having three different lists) makes 
        # it easy to write protected helpers to do most of the work.  

        self._callbacks = {
                'message': [],
                'sync_response': [],
                'undo_response': [],
        }

        # Create a member variable indicating whether or not the ability to 
        # subscribe to or unsubscribe from messages should be enabled.  Token 
        # disable this functionality until they've been added to the world and 
        # RemoteActors disable it permanently. 

        self._is_enabled = True

        # Decorators can be used to automatically label methods that should be 
        # callbacks.  Here, we look for methods that have been labeled in this 
        # way and register them appropriately.

        from inspect import getmembers, ismethod

        for _, method in getmembers(self, ismethod):
            for message_cls in getattr(method, '_kxg_subscribe_to_message', []):
                self.subscribe_to_message(message_cls, method)

            for message_cls in getattr(method, '_kxg_subscribe_to_sync_response', []):
                self.subscribe_to_sync_response(message_cls, method)

            for message_cls in getattr(method, '_kxg_subscribe_to_undo_response', []):
                self.subscribe_to_undo_response(message_cls, method)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_callbacks']
        del state['_is_enabled']
        return state

    def __setstate__(self, state):
        ForumObserver.__init__(self)
        self.__dict__.update(state)

    def subscribe_to_message(self, message_cls, callback):
        self._add_callback('message', message_cls, callback)

    def subscribe_to_sync_response(self, message_cls, callback):
        self._add_callback('sync_response', message_cls, callback)

    def subscribe_to_undo_response(self, message_cls, callback):
        self._add_callback('undo_response', message_cls, callback)

    def unsubscribe_from_message(self, message_cls, callback=None):
        self._drop_callback('message', message_cls, callback)

    def unsubscribe_from_sync_response(self, message_cls, callback=None):
        self._drop_callback('sync_response', message_cls, callback)

    def unsubscribe_from_undo_response(self, message_cls, callback=None):
        self._drop_callback('undo_response', message_cls, callback)

    def _react_to_message(self, message):
        self._call_callbacks('message', message)

    def _react_to_sync_response(self, message):
        self._call_callbacks('sync_response', message)

    def _react_to_undo_response(self, message):
        self._call_callbacks('undo_response', message)

    def _enable_forum_observation(self):
        self._is_enabled = True

    def _disable_forum_observation(self):
        self._is_enabled = False

    def _check_if_forum_observation_enabled(self):
        assert self._is_enabled, "{} has disabled forum observation.".format(self)

    def _add_callback(self, event, message_cls, callback):
        from .messages import require_message_cls
        require_message_cls(message_cls)
        self._check_if_forum_observation_enabled()
        callback_info = ForumObserver.CallbackInfo(message_cls, callback)
        self._callbacks[event].append(callback_info)

    def _drop_callback(self, event, message_cls, callback):
        from .messages import require_message_cls
        require_message_cls(message_cls)
        self._check_if_forum_observation_enabled()
        self._callbacks[event] = [
                x for x in self._callbacks[event]
                if not ((x.message_cls is message_cls) and
                        (x.callback is callback or callback is None)
                )
        ]
        
    def _call_callbacks(self, event, message):
        self._check_if_forum_observation_enabled()

        # Call the callbacks stored in this observer.

        for callback_info in self._callbacks[event]:
            if isinstance(message, callback_info.message_cls):
                callback_info.callback(message)

        # Call the callbacks stored in nested observers.

        for observer in self._get_nested_observers():
            observer._call_callbacks(event, message)

    def _get_nested_observers(self):
        return []


class IdFactory:

    def __init__(self, offset, spacing):
        self.offset = offset
        self.spacing = spacing
        self.num_ids_assigned = 0

    def __repr__(self):
        return 'IdFactory(offset={}, spacing={})'.format(
                self.offset, self.spacing)

    def __contains__(self, id):
        return id % self.spacing == self.offset % self.spacing

    def get(self):
        return self.offset

    def next(self):
        next_id = self.num_ids_assigned * self.spacing + self.offset
        self.num_ids_assigned += 1
        return next_id



@debug_only
def require_forum(object):
    require_instance(Forum(), object)

def subscribe_to_message(message_cls):
    def decorator(function):
        try:
            function._kxg_subscribe_to_message.append(message_cls)
        except AttributeError:
            function._kxg_subscribe_to_message = [message_cls]
        return function
    return decorator

def subscribe_to_sync_response(message_cls):
    def decorator(function):
        try:
            function._kxg_subscribe_to_sync_response.append(message_cls)
        except AttributeError:
            function._kxg_subscribe_to_sync_response = [message_cls]
        return function
    return decorator

def subscribe_to_undo_response(message_cls):
    def decorator(function):
        try:
            function._kxg_subscribe_to_undo_response.append(message_cls)
        except AttributeError:
            function._kxg_subscribe_to_undo_response = [message_cls]
        return function
    return decorator





#!/usr/bin/env python3

from .errors import *
from .forums import ForumObserver

class Actor(ForumObserver):

    def __init__(self):
        super().__init__()
        self.world = None
        self._forum = None
        self._id_factory = None

    def __rshift__(self, message):
        return self.send_message(message)

    def send_message(self, message):
        info("sending a message: {message}")

        # Make sure the user didn't pass the wrong object to this function or 
        # forget to call the superclass constructor.

        from .messages import require_message
        require_message(message)

        # Make sure this message hasn't been sent more than once.  This is 
        # conceptually dangerous because messages can accumulate state as they 
        # are processed, and practically dangerous because it breaks the sent 
        # message cache in multiplayer games.

        if message.was_sent():
            raise CantReuseMessage()

        # Indicate that the message was sent by this actor and give the message 
        # a chance to assign id numbers to the tokens it's creating.  This is 
        # done before the message is checked so that the check can make sure 
        # valid ids were assigned.

        message._set_sender_id(self._id_factory)
        message._assign_token_ids(self._id_factory)

        # Make sure that the message isn't requesting something that can't be 
        # done.  For example, make sure the players have enough resource when 
        # they're trying to buy things.  If the message fails the check, return 
        # False immediately.

        message._check(self.world)

        # Hand the message off to the forum to be applied to the world and 
        # relayed on to all the other actors (which may or may not be on 
        # different machines).

        self._forum.execute_message(message)
        return True

    @property
    def id(self):
        assert self._id_factory is not None, "Actor does not have id."
        return self._id_factory.get()

    def is_referee(self):
        return isinstance(self, Referee)

    def on_setup_gui(self, gui):
        pass

    def on_start_game(self):
        pass

    def on_update_game(self, dt):
        pass    # pragma: no cover

    def on_finish_game(self):
        pass

    def _set_world(self, world):
        assert self.world is None, "can't set world twice"
        self.world = world

    def _set_forum(self, forum, id_factory):
        assert self._id_factory is None, "Actor already has id."
        self._id_factory = id_factory

        assert self._forum is None, "Actor already has forum."
        self._forum = forum

    def _get_nested_observers(self):
        return (token.get_extension(self)
                for token in self.world if token.has_extension(self))

    def _relay_message(self, message):
        pass


class Referee(Actor):

    class Reporter:
        
        def __init__(self, referee):
            self.referee = referee
            self.is_finished_reporting = False

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.is_finished_reporting = True

        def __rshift__(self, message):
            return self.send_message(message)

        def send_message(self, message):
            if self.is_finished_reporting:
                raise UsingStaleReporter(message)
            else:
                return self.referee.send_message(message)

    def on_update_game(self, dt):
        with Referee.Reporter(self) as reporter:
            for token in self.world:
                token.on_report_to_referee(reporter)

    def _set_forum(self, forum, id_factory):
        super()._set_forum(forum, id_factory)
        assert self.id == 1



@debug_only
def require_actor(object):
    require_instance(Actor(), object)

@debug_only
def require_actors(objects):
    for object in objects:
        require_actor(object)



from .errors import *
from .forums import Forum, IdFactory
from .actors import Actor

class ClientForum(Forum):

    def __init__(self, pipe):
        super().__init__()
        self.pipe = pipe
        self.pipe.lock()

        from collections import OrderedDict

        self.actor_id_factory = None
        self.response_id_factory = IdFactory(0, 1)
        self.sent_message_cache = OrderedDict()

    def receive_id_from_server(self):
        """
        Listen for an id from the server.

        At the beginning of a game, each client receives an IdFactory from the 
        server.  This factory are used to give id numbers that are guaranteed 
        to be unique to tokens that created locally.  This method checks to see if such 
        a factory has been received.  If it hasn't, this method does not block 
        and immediately returns False.  If it has, this method returns True 
        after saving the factory internally.  At this point it is safe to enter 
        the GameStage.
        """
        for message in self.pipe.receive():
            if isinstance(message, IdFactory):
                self.actor_id_factory = message
                return True
        return False

    def connect_everyone(self, world, actors):
        # Make sure that this forum is only connected to one actor.

        assert len(actors) == 1
        self.actor = actors[0]

        # Connect the forum, world, and actors as usual.

        super().connect_everyone(world, actors)

    def execute_message(self, message):
        # Cache the message and give it an id number the server can reference 
        # in its response.  Messages are cached so they can be undone if they 
        # are rejected by the server.  The id is necessary so the client forum 
        # (i.e. this object) can associate each response with a cached message.

        message._set_server_response_id(self.response_id_factory.next())
        self.sent_message_cache[message._get_server_response_id()] = message

        # Relay the message to a ServerActor running on the server to update 
        # the world on all of the other machines playing the game as well.

        self.pipe.send(message)
        self.pipe.deliver()

        # Have the message update the local world like usual.

        super().execute_message(message)

    def execute_sync(self, message):
        """
        Respond when the server indicates that the client is out of sync.

        The server can request a sync when this client sends a message that 
        fails the check() on the server.  If the reason for the failure isn't 
        very serious, then the server can decide to send it as usual in the 
        interest of a smooth gameplay experience.  When this happens, the 
        server sends out an extra response providing the clients with the
        information they need to resync themselves.
        """
        info("synchronizing a message: {message}")

        # Synchronize the world.

        with self.world._unlock_temporarily():
            message._sync(self.world)
            self.world._react_to_sync_response(message)

        # Synchronize the tokens.

        for actor in self.actors:
            actor._react_to_sync_response(message)

    def execute_undo(self, message):
        """
        Manage the response when the server reports a hard sync error.

        A hard sync error is produced when this client sends a message that the 
        server refuses to pass on to the other clients playing the game.  In 
        this case, the client must either undo the changes that the message 
        made to the world before being sent or crash.  Note that unlike a soft 
        sync error, a hard sync error is only reported to the client that sent 
        the offending message.
        """
        info("undoing a message: {message}")

        # Roll back changes that the original message made to the world.

        with self.world._unlock_temporarily():
            message._undo(self.world)
            self.world._react_to_undo_response(message)

        # Give the actors a chance to react to the error.  For example, a 
        # GUI actor might inform the user that there are connectivity 
        # issues and that their last action was countermanded.

        for actor in self.actors:
            actor._react_to_undo_response(message)

    def on_start_game(self):
        from .tokens import TokenSerializer
        serializer = TokenSerializer(self.world)
        self.pipe.push_serializer(serializer)

    def on_update_game(self):
        from .messages import Message

        # An attempt is made to immediately deliver any messages passed into 
        # execute_message(), but sometimes it takes more than one try to send a 
        # message.  So in case there are any messages waiting to be sent, the 
        # code below attempts to clear the queue every frame.

        self.pipe.deliver()

        # For each message received from the server:

        for packet in self.pipe.receive():

            # If the incoming packet is a message, execute it on this client 
            # and, if necessary, synchronize this client's world with the 
            # server's.  Messages that were sent from this client will not 
            # reappear here, so we don't need to worry about double-dipping.

            if isinstance(packet, Message):
                info("receiving a message: {packet}")
                super().execute_message(packet)
                response = packet._get_server_response()
                if response and response.sync_needed:
                    self.execute_sync(packet)

            # If the incoming packet is a response to a message sent from this 
            # client, find that message in the "sent message cache" and attach 
            # the response to it.  The response is handled in the while loop 
            # below (and not right here) to better handle weird cases that can 
            # occur when several messages are sent between server responses.

            elif isinstance(packet, ServerResponse):
                message = self.sent_message_cache[packet.id]
                message._set_server_response(packet)

        # Try to clear the sent message cache:

        while self.sent_message_cache:
            message = self.sent_message_cache[next(reversed(self.sent_message_cache))]
            response = message._get_server_response()

            # Don't handle any response until responses for any messages that 
            # were sent after it have been handled.  This keeps the world in a 
            # sane state for every response.

            if response is None:
                break

            # If the server requested that a message sync or undo itself, then 
            # do that.  Messages coming from any client may need to be synced, 
            # but messages that need to be undone were sent by this client and 
            # rejected by the server.

            if response.sync_needed:
                self.execute_sync(message)
            if response.undo_needed:
                self.execute_undo(message)

            # Now that the message has been fully handled, pop it off the 
            # cache.

            self.sent_message_cache.popitem()

    def on_finish_game(self):
        self.pipe.pop_serializer()

    def _assign_id_factories(self):
        assert self.actor_id_factory is not None
        return {self.actor: self.actor_id_factory}


class ServerActor(Actor):

    def __init__(self, pipe):
        super().__init__()
        self._disable_forum_observation()
        self.pipe = pipe
        self.pipe.lock()

    def send_message(self, message):
        raise NotImplementedError

    def on_start_game(self):
        from .tokens import TokenSerializer
        serializer = TokenSerializer(self.world)
        self.pipe.push_serializer(serializer)

    def on_update_game(self, dt):
        from .messages import MessageCheck

        # For each message received from the connected client:

        for message in self.pipe.receive():
            info("received a message: {message}")

            # Make sure the message wasn't sent by an actor with a different id 
            # than this one.  This should absolutely never happen because this 
            # actor gives its id to its client, so if a mismatch is detected 
            # there's probably a bug in the game engine.

            if not message.was_sent_by(self._id_factory):
                critical("ignoring message from player {self.id} claiming to be from player {message.sender_id}.")
                continue

            # Check the message to make sure it matches the state of the game 
            # world on the server.  If the message doesn't pass the check, the 
            # client and server must be out of sync, because the same check was 
            # just passed on the client.

            response = ServerResponse(message)
            try:
                message._check(self.world)
            except MessageCheck:
                response.sync_needed = True
            else:
                response.sync_needed = False

            # Decide if it will be enough for the clients to sync themselves, 
            # or if this message shouldn't be relayed at all (and should be 
            # undone on the client that sent it).  The message is also given a 
            # chance to store information it can use later to sync the game. 

            if response.sync_needed:
                response.undo_needed = not message._prepare_sync(
                        self.world, response)

            # Tell the clients how to treat this message.  For the client that 
            # sent the message in the first place, the response is sent on its 
            # own.  If a sync or an undo is needed, the client will retrieve 
            # the original message from its cache and use it to reconcile its 
            # world with the server's.  Otherwise, the client will just clear 
            # the original message from its cache.  For all the other clients, 
            # the response is attached to the message, but only if a sync is 
            # needed (otherwise nothing special needs to be done).

            self.pipe.send(response)

            # If the message doesn't have an irreparable sync error, execute it 
            # on the server and relay it to all the other clients.

            if not response.undo_needed:
                self._forum.execute_message(message)

        # Deliver any messages waiting to be sent.  This has to be done every 
        # frame because it sometimes takes more than one try to send a message.

        self.pipe.deliver()

    def on_finish_game(self):
        self.pipe.pop_serializer()

    def _set_forum(self, forum, id):
        super()._set_forum(forum, id)
        self.pipe.send(id)

    def _relay_message(self, message):
        """
        Relay messages from the forum on the server to the client represented 
        by this actor.
        """
        info("relaying a message: {message}")

        if not message.was_sent_by(self._id_factory):
            self.pipe.send(message)
            self.pipe.deliver()

    def _react_to_message(self, message):
        """
        Don't ever change the world in response to a message.

        This method is defined is called by the game engine to trigger 
        callbacks tied by this actor to particular messages.  This is useful 
        for ordinary actors, but remote actors are only meant to shuttle 
        message between clients and should never react to individual messages.
        """
        pass


class ServerResponse:

    def __init__(self, message):
        self.id = message._get_server_response_id()
        self.sync_needed = False
        self.undo_needed = False

    def __repr__(self):
        return "{}(sync_needed={}, undo_needed={})".format(
                self.__class__.__name__, self.sync_needed, self.undo_needed)
    




#!/usr/bin/env python3
# pylint: disable=unused-import

from nonstdlib import log, debug, info, warning, error, critical

class ApiUsageError(Exception):
    """
    A class for communicating errors to the user.
    """

    message_width = 79

    def __init__(self, message):
        import re, textwrap

        width = ApiUsageError.message_width
        message = textwrap.dedent(message)
        tokens = [x.strip() for x in re.split(r'\n\s*\n', message, 1)]

        # Make sure the summary doesn't overflow 80 characters even after 
        # python adds the 'kxg.errors.ApiUsageError: ' prefix.

        prefix = (
                self.__class__.__module__ + '.' +
                self.__class__.__name__ + ': '
        )
        summary = textwrap.fill(
                tokens.pop(0), width=width, initial_indent=' ' * len(prefix),
        ).strip()

        # If a details paragraph was given, wrap it to fit in 80 characters.

        if tokens:
            details = tokens.pop(0).replace('\n', '')
            details = '\n\n' + textwrap.fill(details, width=width)
        else:
            details = ''

        super().__init__(summary + details)


class ApiUsageErrorFactory:
    """
    Factory class for making nicely formatted exceptions.

    ApiUsageError is the only kind of exception raised by the game engine.  
    Different kinds of errors are distinguished by different error messages, 
    but not by different classes.  The reason for this is mostly aesthetic: 
    every message prints out a brief sentence describing the problem, and this 
    sentence makes a specific class name redundant.  Specific error class names 
    can also be quite long.  

    However, the engine still needs an easy way to raise API usage errors with 
    different error messages.  The solution to this problem is this class: 
    ApiUsageErrorFactory.  This class is meant to be subclassed.  Each subclass 
    represents a single type of error and knows both how to describe that error 
    and when to raise that error.

    To define a new type of error, start by making a new ApiUsageErrorFactory 
    subclass.  You can overload the get_message() method to return a message 
    describing your error.  You can also use the default implementation, which 
    is basically just 'return self.message.format(**self.__dict__)'.  In other 
    words, you can just provide an attribute called 'message' attribute which 
    may contain formatting references to any other attribute of your class 
    (e.g. "Token {token} has a bad id." if your class has a 'token' attribute.)  
    This default approach should work well for most cases.

    You error message should be composed of a one-line summary and a multi-line 
    description.  The summary and the description should be separated by a 
    blank line.  This format is parsed by ApiUsageError so it can format the 
    message nicely.  The summary should simple state what went wrong.  The 
    description should elaborate on that and should suggest the most likely 
    causes and solutions.

    You can also implement the error_condition() method to return whether or 
    not the error in question should be raised.  This method is used by the 
    raise_if_warranted() method to decide whether or not the exception should 
    be raised.  I also think it's good practice to keep description of the 
    error and the logic for raising the error in the same place.

    Note that when you try to construct an instance of an ApiUsageErrorFactory 
    subclass, you are actually returned an ApiUsageError instance.  So:

    >>> class SirRobinRanAway(ApiUsageErrorFactory): pass
    >>> type(SirRobinRanAway())
    <class 'kxg.ApiUsageError'>

    This is accomplished using some python magic involving __new__() and is 
    meant to cut down on boilerplate.  But I can also see it being confusing 
    if you're looking at things closely.  The best way to think about this is 
    to remember that your factory classes are really factories, and that the 
    constructor is the when the exception is created.
    """

    message = "Unknown error."

    def __new__(cls, *args, **kwargs):
        factory = object.__new__(cls)
        factory.__init__(*args, **kwargs)
        message = factory.message.format(**factory.__dict__)

        # Normally, we want to lowercase the first letter in every error 
        # message to force adherence to the recommended python style.  The 
        # exception is if the first character in the unformatted message is a 
        # brace, which means that the message probably begins with some 
        # identifier name.  In that case we don't want to change anything.

        if not factory.message.startswith('{'):
            message = message[0].lower() + message[1:]

        return ApiUsageError(message)



class CantModifyTokenIfWorldLocked(ApiUsageErrorFactory):

    message = """\
attempted unsafe invocation of {token_class}.{method_name}().

This error brings attention to situations that might cause synchronization 
issues in multiplayer games.  The {method_name}() method is not marked as 
read-only, but it was invoked from outside the context of a message.  This 
means that if {method_name}() makes any changes to the world, those changes 
will not be propagated.  If {method_name}() is actually read-only, label it 
with the @kxg.read_only decorator."""

    def __init__(self, token, method_name):
        self.token = token
        self.token_class = token.__class__.__name__
        self.method_name = method_name


class CantModifyTokenIfNotInWorld(ApiUsageErrorFactory):

    message = """\
may have forgotten to add {token} to the world.

The {token_class}.{method_name}() method was invoked on a token that had not 
yet been added to the game world.  This is usually a sign that you forgot to 
add the token in question was to the game world.  Label the {method_name}() 
method with the @kxg.before_world decorator if you really do need to call it 
before the token has been added to the world (i.e. the method helps setup 
{token_class} tokens)."""

    def __init__(self, token, method_name):
        self.token = token
        self.token_class = token.__class__.__name__
        self.method_name = method_name


class CantPickleWorld(ApiUsageErrorFactory):

    message = """\
can't pickle the world.

The world should never have to be pickled and sent over the network, because 
each machine starts with its own world and is kept in sync by the messaging 
system.  But unless you are explicitly trying to pickle the world on your own, 
this error is more likely to be the symptom of a major bug in the messaging 
system that is preventing it from correctly deciding which tokens need to be 
pickled."""

class CantResetActiveToken(ApiUsageErrorFactory):

    message = """\
token {token} can't be reset because it's still being used.

The purpose of Token.reset_registration() is to allow you to reuse tokens that 
were previously removed from the world.  Normally these tokens aren't allowed 
to do anything, which makes it easy to know when you accidentally keep a stale 
reference to a token that isn't part of the world anymore.  This error means 
that you tried to reset a token that was never removed from the world (or that 
was already reset)."""

    def __init__(self, token):
        self.token = token


class CantReuseMessage(ApiUsageErrorFactory):

    message = """\
can't send the same message more than once.

It's not safe to send the same message twice because messages can accumulate 
state as they are executed.  This also breaks the system by which clients can 
react to responses from the server in multiplayer games.
"""
class IllegalTokenExtensionConstructor(ApiUsageErrorFactory):

    message = """\
the {extension_cls} constructor doesn't take the right arguments.

Token extension constructors must take exactly three arguments: self, actor, 
and token.  These are the arguments provided by tokens when they automatically 
instantiates their extensions.  Fix this error by making the {extension_cls} 
constructor compatible with these arguments."""

    def __init__(self, extension_cls):
        self.extension_cls = extension_cls.__name__


class MessageNotYetSent(ApiUsageErrorFactory):

    message = """\
can't ask who sent a message before it's been sent.

This error means Message.was_sent_by() or Message.was_sent_by_referee() got 
called on a message that hadn't been sent yet.  Normally you would only call 
these methods from within Message.on_check()."""

class MessageAlreadySent(ApiUsageErrorFactory):

    message = """\
messages can't add or remove tokens after they've been sent.

You get this error if you call Message.add_token() or Message.remove_token() 
after the message has been sent.  Normally you would only call these methods 
from within the constructors of your message subclasses."""

class NotUsingIdFactory(ApiUsageErrorFactory):

    message = """\
can't use {bad_id} as a token id.

Token._give_id() expects to be passed an internal object that can be used to 
create new, unique id numbers.  For that reason, this method should only be 
called by the game engine.  You can get this error if you try to assign an id 
to a token by calling Token._give_id() yourself.  Use Message.add_token() to 
create all your tokens in a manner that avoids synchronization bugs."""

    def __init__(self, bad_id):
        self.bad_id = bad_id


class ObjectIsntRightType(ApiUsageErrorFactory):

    message = """\
expected {prototype_cls}, but got {object_cls} instead."""

    def __init__(self, prototype, object):
        self.prototype = prototype
        self.prototype_cls = prototype.__class__.__name__
        self.object = object
        self.object_cls = object.__class__.__name__


class ObjectIsntFullyConstructed(ApiUsageErrorFactory):

    message = """\
forgot to call the {prototype_cls} constructor in {object_cls}.__init__().

The game engine was passed an object that inherits from {prototype_cls} but is 
missing the '{missing_member}' attribute.  This usually means that you forgot 
to call the {prototype_cls} constructor in your subclass."""

    def __init__(self, prototype, object, missing_member):
        self.prototype = prototype
        self.prototype_cls = prototype.__class__.__name__
        self.object = object
        self.object_cls = object.__class__.__name__
        self.missing_member = missing_member


class ObjectIsntMessageSubclass(ApiUsageErrorFactory):

    message = """\
expected Message subclass, but got {object} instead."""

    def __init__(self, object):
        try:
            self.object = object.__name__
        except:
            self.object = object


class TokenAlreadyHasId(ApiUsageErrorFactory):

    message = """\
token {token} already has an id.

This error usually means that {token} was added to the world twice."""

    def __init__(self, token):
        self.token = token


class TokenAlreadyInWorld(ApiUsageErrorFactory):

    message = """\
can't add the same token to the world twice.

Token {token} can't be added to the world, because the world already contains a 
token with the id={token.id}.  You can get this error if you try to add a token 
to the world on without using a message (i.e. CreateToken)."""

    def __init__(self, token):
        self.token = token


class TokenDoesntHaveId(ApiUsageErrorFactory):

    message = """\
token {token} should have an id, but doesn't.

This error usually means that a token was added to the world without being 
assigned an id number.  To correct this, make sure that you're using a message 
(i.e. CreateToken) to create all of your tokens."""

    def __init__(self, token):
        self.token = token


class TokenCantSubscribeNow(ApiUsageErrorFactory):

    message = """\
token {token} can't subscribe to messages now.

Tokens must be registered with the world before they can subscribe to (or 
unsubscribe from) messages, because subscriptions can't be pickled and sent 
over the network.  So any subscriptions a token makes while it's not part of 
the world won't be communicated to each machine playing the game.  You are most 
likely getting this error because you tried to subscribe to messages in the 
constructor of a Token subclass.  You can't do that, but instead you can either 
make your subscriptions in the on_add_to_world() callback or you can label your 
handler methods with the @subscribe_to_message decorator."""

    def __init__(self, token):
        self.token = token


class TokenCantUseStaleReporter(ApiUsageErrorFactory):
    
    message = """\
tokens can't send messages outside of report().

The engine detected a '{message.__class__.__name__}' message being sent to a 
reporter that is no longer accepting messages.  This can happen if you save the 
reporter object passed to Token.report() and attempt to use it outside of that 
method call."""

class TokenHasNoSuchMethodToWatch(ApiUsageErrorFactory):

    message = """\
{token_cls} has no such method {method_name}() to watch.

This error usually means that you used the @watch_token decorator on a method 
of a token extension class that didn't match the name of any method in the 
corresponding token class.  Check for typos."""

    def __init__(self, token, method_name):
        self.token = token
        self.token_cls = token.__class__.__name__
        self.method_name = method_name


class TokenNotInWorld(ApiUsageErrorFactory):

    message = """\
token {token} (id={token.id}) not in world.

You can get this error if you try to remove the same token from the world 
twice.  This might happen is you don't get rid of every reference to a token 
after it's removed the first time, then later on you try to remove the stale 
reference."""

    def __init__(self, token):
        self.token = token


class UsingRemovedToken(ApiUsageErrorFactory):

    message = """\
token {token} has been removed from the world.

This error is raised when the game engine detects that a token that has been 
removed from world is still being used.  For example, maybe a message can't be 
pickled because it contains a reference to a token that was previously removed 
from the world.  This error usually means that a stale reference to a destroyed 
token is accidentally being kept."""

    def __init__(self, token):
        self.token = token


class UsingStaleReporter(ApiUsageErrorFactory):

    message = """\
{message_cls} message sent using a stale reporter.

This message is raised when the reporter provided to Token.report() is used 
after that method returns.  This can only happen if you save a reference to the 
reporter, which you shouldn't do.  This is a multiplayer synchronization issue.  
Because the same token should exist in the same state on every machine playing 
the game, tokens normally can't send messages or they would be prone to sending 
duplicate messages.  The exception is Token.report(), which is guaranteed to be 
called only on the server.  Token.report() is provided with a reporter object 
that can be used to send messages, but it's illegal to save a reference to the 
reporter and use it after Token.report() returns.  Such a design would lead to 
bugs on the clients, which are never given reporter objects."""

    def __init__(self, message):
        self.message_cls = message.__class__.__name__


class UnhandledSyncError(ApiUsageErrorFactory):

    message = """\
the message {message_} was rejected by the server.

This client attempted to send a {message_cls} message, but it was rejected by 
the server.  To fix this error, either figure out why the client is getting out 
of sync with the server or implement a {message_cls}.on_undo() that undoes 
everything done in {message_cls}.on_execute()."""

    def __init__(self, message):
        self.message_ = message
        self.message_cls = message.__class__.__name__


class WorldAlreadyUnlocked(ApiUsageErrorFactory):

    message = """\
tired to unlock the world, but it's already unlocked.

You can't get this error unless you manually call World._unlock_temporarily(), 
which you should never do.  This method is intended to be used by the game 
engine, which was carefully designed to allow the world to be modified only 
when safe.  Calling this method yourself disables an important safety check.
"""

def debug_only(function):
    if __debug__:
        return function
    else:
        return lambda *args, **kwargs: None

@debug_only
def require_instance(prototype, object):
    if not isinstance(object, type(prototype)):
        raise ObjectIsntRightType(prototype, object)

    for member_name in prototype.__dict__:
        if not hasattr(object, member_name):
            raise ObjectIsntFullyConstructed(prototype, object, member_name)

    return object




#!/usr/bin/env python3

import contextlib

from .errors import *
from .tokens import Token, read_only, require_token, require_active_token

class World(Token):

    def __init__(self):
        super().__init__()
        self._id = 0
        self._tokens = {}
        self._actors = []
        self._is_locked = True
        self._is_game_over = False
        with self._unlock_temporarily():
            self._add_token(self)

    def __repr__(self):
        return '{}(len={})'.format(self.__class__.__name__, len(self))

    def __iter__(self):
        return (x for x in self._tokens.values() if x is not self)

    def __len__(self):
        return len(self._tokens)

    def __contains__(self, token_or_id):
        id = token_or_id.id if isinstance(token_or_id, Token) else token_or_id
        return id in self._tokens

    def __getstate__(self):
        raise CantPickleWorld()

    def __setstate__(self, state):
        raise CantPickleWorld()     # pragma: no cover

    @read_only
    def get_token(self, id):
        """
        Return the token with the given id.  If no token with the given id is 
        registered to the world, an IndexError is thrown.
        """
        return self._tokens[id]

    @read_only
    def get_last_id(self):
        """
        Return the largest token id registered with the world.  If no tokens 
        have been added to the world, the id for the world itself (0) is 
        returned.  This means that the first "real" token id is 1.
        """
        return max(self._tokens)

    @read_only
    def is_locked(self):
        """
        Return whether or not the world is currently allowed to be modified.
        """
        return self._is_locked

    @read_only
    def is_game_over(self):
        """
        Return true if the game has ended.
        """
        return self._is_game_over

    def end_game(self):
        self._is_game_over = True

    def on_start_game(self):
        pass

    def on_update_game(self, dt):
        for token in self:
            token.on_update_game(dt)

    def on_finish_game(self):
        pass

    @contextlib.contextmanager
    def _unlock_temporarily(self):
        """
        Allow tokens to modify the world for the duration of a with-block.

        It's important that tokens only modify the world at appropriate times, 
        otherwise the changes they make may not be communicated across the 
        network to other clients.  To help catch and prevent these kinds of 
        errors, the game engine keeps the world locked most of the time and 
        only briefly unlocks it (using this method) when tokens are allowed to 
        make changes.  When the world is locked, token methods that aren't 
        marked as being read-only can't be called. only allows token methods 
        When the world is unlocked, any token method can be called.  These 
        checks can be disabled by running python with optimization enabled.

        You should never call this method manually from within your own game.  
        This method is intended to be used by the game engine, which was 
        carefully designed to allow the world to be modified only when safe.  
        Calling this method yourself disables an important safety check.
        """
        if not self._is_locked:
            yield
        else:
            try:
                self._is_locked = False
                yield
            finally:
                self._is_locked = True

    def _add_token(self, token):
        require_token(token)
        if token.world_registration == 'pending' and not token.has_id():
            raise TokenDoesntHaveId(token)
        if token.world_registration == 'active':
            raise TokenAlreadyInWorld(token)
        if token.world_registration == 'expired':
            raise UsingRemovedToken(token)

        # Add the token to the world.  This means adding it to the world's list 
        # of tokens, giving it a reference to the world (which allows it to 
        # use methods that haven't been marked with @before_setup), and 
        # allowing it to subscribe to messages from the forum.

        id = token.id
        self._tokens[id] = token
        token._world = self
        token._removed_from_world = False
        token._enable_forum_observation()

        # Initialize any extensions relevant to this token.  Which extensions 
        # are relevant depends on which actors are being are running on the 
        # current machine.  At most one extension will be created per actor.

        token._extensions = {}
        extension_classes = token.__extend__()

        for actor in self._actors:
            actor_class = type(actor)
            extension_class = extension_classes.get(actor_class)

            if extension_class:

                # Make sure the extension class constructor takes exactly three 
                # arguments: self, actor, token.  If the constructor has a 
                # different signature, raise an error with a helpful message.

                from inspect import getargspec
                argspec = getargspec(extension_class.__init__)
                if len(argspec.args) != 3:
                    raise IllegalTokenExtensionConstructor(extension_class)

                # Instantiate the extension and store a reference to it.

                extension = extension_class(actor, token)
                token._extensions[actor] = extension

        # Finally, give the token a chance to react to it's own creation.

        token.on_add_to_world(self)

        return token

    def _remove_token(self, token):
        require_active_token(token)

        # Remove the token from the world.  The token is given a chance to 
        # react to its destruction, then forum observation is disabled.

        token.on_remove_from_world()
        token._disable_forum_observation()
        token._extensions = {}
        token._removed_from_world = True
        token._world = None

        id = token.id
        del self._tokens[id]

        # Don't reset the token's id (e.g. token._id = None) because we might 
        # need to undo this action and re-add the token to the world (e.g. if 
        # the server rejects the message that triggered this call).  In order 
        # to undo that action we need to know the token's original id number.

    def _get_nested_observers(self):
        return iter(self)

    def _set_actors(self, actors):
        """
        Tell the world which actors are running on this machine.  This 
        information is used to create extensions for new tokens.  
        """
        self._actors = actors



@debug_only
def require_world(object):
    return require_instance(World(), object)




#!/usr/bin/env python3

import time, pyglet
from .errors import *
from .forums import Forum
from .multiplayer import ClientForum, ServerActor 

class Theater:
    """
    Manage whichever stage is currently active.  This involves both
    updating the current stage and handling transitions between stages.
    """

    def __init__(self, initial_stage=None, gui=None):
        self._gui = gui
        self._initial_stage = initial_stage
        self._current_stage = None
        self._current_update = self._update_before_loop
        self._previous_time = time.time()

    @property
    def gui(self):
        return self._gui

    @gui.setter
    def gui(self, gui):
        if self._current_stage:
            raise ApiUsageError("theater already playing; can't set gui.")
        self._gui = gui

    @property
    def initial_stage(self):
        return self._initial_stage

    @initial_stage.setter
    def initial_stage(self, stage):
        require_stage(stage)
        if self._current_stage:
            raise ApiUsageError("theater already playing; can't set initial stage.")
        self._initial_stage = stage

    @property
    def current_stage(self):
        return self._current_stage

    @property
    def is_finished(self):
        return self._current_update == self._update_after_loop

    def update(self, dt=None):
        self._current_update(dt)

    def exit(self):
        if self._current_stage:
            self._current_stage.on_exit_stage()

        self._current_update = self._update_after_loop


    def _update_before_loop(self, dt):
        self._current_stage = self._initial_stage
        self._current_stage.theater = self
        self._current_stage.on_enter_stage()

        self._current_update = self._update_main_loop
        self._current_update(dt)

    def _update_main_loop(self, dt):
        if dt is None:
            current_time = time.time()
            dt = current_time - self._previous_time
            self._previous_time = current_time

        self._current_stage.on_update_stage(dt)

        if self._current_stage.is_finished:
            self._current_stage.on_exit_stage()
            self._current_stage = self._current_stage.successor

            if not self._current_stage:
                self.exit()
            else:
                require_stage(self._current_stage)
                self._current_stage.theater = self
                self._current_stage.on_enter_stage()

    def _update_after_loop(self, dt):
        raise AssertionError("shouldn't update the theater after the game has ended.")


class PygletTheater(Theater):

    def play(self, frames_per_sec=50):
        pyglet.clock.schedule_interval(self.update, 1/frames_per_sec)
        pyglet.app.run()

    def exit(self):
        super().exit()
        pyglet.app.exit()


class Stage:

    def __init__(self):
        self.theater = None
        self.successor = None
        self.is_finished = False

    @property
    def gui(self):
        return self.theater.gui

    def exit_stage(self):
        """
        Stop this stage from executing once the current update ends.
        """
        self.is_finished = True

    def exit_theater(self):
        """
        Exit the game once the current update ends.
        """
        self.theater.exit()

    def on_enter_stage(self):
        """
        Give the stage a chance to set itself up before it is updated for the 
        first time.
        """
        pass

    def on_update_stage(self, dt):
        """
        Give the stage a chance to react to each clock cycle.

        The amount of time that passed since the last clock cycle is provided 
        as an argument.
        """
        pass

    def on_exit_stage(self):
        """
        Give the stage a chance to react before it is stopped and the next 
        stage is started.
        
        You can define the next stage by setting the Stage.successor attribute.  
        If the successor is static, you can just set it in the constructor.  
        But if it will differ depending on the context, this method may be a 
        good place to calculate it because it is called only once and just 
        before the theater queries for the successor.
        """
        pass


class GameStage(Stage):

    def __init__(self, world, forum, actors):
        Stage.__init__(self)

        from .world import require_world; require_world(world)
        from .forums import require_forum; require_forum(forum)
        from .actors import require_actors; require_actors(actors)

        self.world = world
        self.forum = forum
        self.actors = actors
        self.successor = None

    def on_enter_stage(self):
        """
        Prepare the actors, the world, and the messaging system to begin 
        playing the game.
        
        This method is guaranteed to be called exactly once upon entering the 
        game stage.
        """
        with self.world._unlock_temporarily():
            self.forum.connect_everyone(self.world, self.actors)

        # 1. Setup the forum.

        self.forum.on_start_game()

        # 2. Setup the world.

        with self.world._unlock_temporarily():
            self.world.on_start_game()

        # 3. Setup the actors.  Because this is done after the forum and the  
        #    world have been setup, this signals to the actors that they can 
        #    send messages and query the game world as usual.

        for actor in self.actors:
            actor.on_setup_gui(self.gui)

        for actor in self.actors:
            actor.on_start_game()

    def on_update_stage(self, dt):
        """
        Sequentially update the actors, the world, and the messaging system.  
        The theater terminates once all of the actors indicate that they are done.
        """

        for actor in self.actors:
            actor.on_update_game(dt)

        self.forum.on_update_game()

        with self.world._unlock_temporarily():
            self.world.on_update_game(dt)

        if self.world.is_game_over():
            self.exit_stage()

    def on_exit_stage(self):
        """
        Give the actors, the world, and the messaging system a chance to react 
        to the end of the game.
        """

        # 1. Let the forum react to the end of the game.  Local forums don't 
        #    react to this, but remote forums take the opportunity to stop 
        #    trying to extract tokens from messages.

        self.forum.on_finish_game()

        # 2. Let the actors react to the end of the game.

        for actor in self.actors:
            actor.on_finish_game()

        # 3. Let the world react to the end of the game.

        with self.world._unlock_temporarily():
            self.world.on_finish_game()


class UniplayerGameStage(GameStage):

    def __init__(self, world, referee, gui_actor, ai_actors=None):
        forum = Forum()
        actors = [referee, gui_actor] + list(ai_actors or [])
        GameStage.__init__(self, world, forum, actors)


class MultiplayerClientGameStage(Stage):

    def __init__(self, world, gui_actor, pipe):
        super().__init__()
        self.forum = ClientForum(pipe)
        self.successor = GameStage(world, self.forum, [gui_actor])

    def on_update_stage(self, dt):
        if self.forum.receive_id_from_server():
            self.exit_stage()


class MultiplayerServerGameStage(GameStage):

    def __init__(self, world, referee, ai_actors, pipes):
        forum = Forum()
        actors = [referee] + [ServerActor(x) for x in pipes] + ai_actors
        GameStage.__init__(self, world, forum, actors)



@debug_only
def require_stage(object):
    require_instance(Stage(), object)





#!/usr/bin/env python3

from .errors import *

class Message:
    # This class defers initializing all of its members until the appropriate 
    # setter is called, rather than initializing everything in a constructor.  
    # This is done to avoid sending unnecessary information over the network.

    class ErrorState:
        SOFT_SYNC_ERROR = 0
        HARD_SYNC_ERROR = 1


    def __repr__(self):
        return self.__class__.__name__ + '()'

    def was_sent(self):
        return hasattr(self, 'sender_id')

    def was_sent_by(self, actor_or_id):
        from .actors import Actor
        from .forums import IdFactory

        if isinstance(actor_or_id, Actor):
            id = actor_or_id.id
        elif isinstance(actor_or_id, IdFactory):
            id = actor_or_id.get()
        else:
            id = actor_or_id

        try:
            return self.sender_id == id
        except AttributeError:
            raise MessageNotYetSent()

    def was_sent_by_referee(self):
        return self.was_sent_by(1)

    def tokens_to_add(self):
        yield from []

    def tokens_to_remove(self):
        yield from []

    def on_check(self, world):
        # Called by the actor.  If no MessageCheck exception is raised, the 
        # message will be sent as usual.  Otherwise, the behavior will depend 
        # on what kind of actor is handling the message.  Actor (uniplayer and 
        # multiplayer clients) will Normal Actor will simply not send the 
        # message.  ServerActor (multiplayer server) will decide if the error 
        # should be handled by undoing the message or asking the clients to 
        # sync themselves.
        raise NotImplementedError

    def on_prepare_sync(self, world, memento):
        # Called only by ServerActor if on_check() returns False.  If this 
        # method returns True, the message will be relayed to the rest of the 
        # clients with the sync error flag set.  Otherwise the message will not 
        # be sent and the ClientForum that sent the message will be instructed 
        # to undo it.  If a soft error is detected, this method should save 
        # information about the world that it could use to resynchronize all 
        # the clients.
        return False

    def on_execute(self, world):
        # Called by the forum on every machine running the game.  Allowed to 
        # make changes to the game world, but should not change the message 
        # itself.  Called before any signal-handling callbacks.
        pass

    def on_sync(self, world, memento):
        # Called by the forum upon receiving a message with the soft error flag 
        # set.  This flag indicates that the client that sent the message is 
        # slightly out of sync with the server, but that the message will be 
        # relayed as usual and that the clients should use the opportunity to 
        # quietly resynchronize themselves.  
        pass

    def on_undo(self, world):
        # Called by ClientForum only upon receiving a message with the hard 
        # error flag set.  This flag indicates that the server refused to relay 
        # the given message to the other clients, presumably because it was too 
        # far out of sync with the world on the server, and that the message 
        # needs to be undone on this client.  Only the ClientForum that sent 
        # the offending message will call this method.
        raise UnhandledSyncError(self)

    def _set_sender_id(self, id_factory):
        self.sender_id = id_factory.get()

    def _set_server_response_id(self, id):
        self._server_response_id = id

    def _get_server_response_id(self):
        return self._server_response_id

    def _set_server_response(self, server_response):
        self._server_response = server_response

    def _get_server_response(self):
        try:
            return self._server_response
        except AttributeError:
            return None

    def _assign_token_ids(self, id_factory):
        # Called by Actor but not by ServerActor, so it is guaranteed to be 
        # called exactly once.  Not really different from the constructor, 
        # except that the id_factory object is nicely provided.  That's useful 
        # for adding tokens but probably nothing else.  This method is called 
        # before _check() so that _check() can make sure that valid ids were 
        # assigned.

        for token in self.tokens_to_add():
            token._give_id(id_factory)

    def _check(self, world):
        self.on_check(world)

    def _prepare_sync(self, world, server_response):
        self._set_server_response(server_response)
        return self.on_prepare_sync(world, self._server_response)

    def _execute(self, world):
        # Deal with tokens to be created or destroyed.

        for token in self.tokens_to_add():
            world._add_token(token)

        for token in self.tokens_to_remove():
            world._remove_token(token)

        # Let derived classes execute themselves.

        self.on_execute(world)

    def _sync(self, world):
        self.on_sync(world, self._server_response)

    def _undo(self, world):
        # The tokens in self.tokens_to_add() haven't been added to the world 
        # yet, because the message was copied and pickled before it was 
        # executed on the server.  We need to access the tokens that are 
        # actually in the world before we can remove them again.

        for token in self.tokens_to_add():
            real_token = world.get_token(token.id)
            world._remove_token(real_token)

        # The tokens in self.tokens_to_remove() have already been removed from 
        # the world.  We want to add them back, and we want to make sure they 
        # end up with the id as before.

        for token in self.tokens_to_remove():
            old_id = token.id
            token.reset_registration()
            token._id = old_id
            world._add_token(token)

        # Let derived classes execute themselves.

        self.on_undo(world)


class MessageCheck(Exception):
    pass


@debug_only
def require_message(object):
    require_instance(Message(), object)

@debug_only
def require_message_cls(cls):
    if not isinstance(cls, type) or not issubclass(cls, Message):
        raise ObjectIsntMessageSubclass(cls)




#!/usr/bin/env python3

import linersock
import multiprocessing, queue
import logging, logging.handlers
from .errors import *
from .theater import *

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 53351

class ClientConnectionStage(Stage):

    def __init__(self, world, gui_actor, host, port):
        super().__init__()
        self.world = world
        self.gui_actor = gui_actor
        self.host = host
        self.port = port
        self.pipe = None
        self.client = linersock.Client(
                host, port, callback=self.on_connection_established)

    def on_update_stage(self, dt):
        self.client.connect()
        try:
            self.gui.on_refresh_gui()
        except AttributeError:
            pass

    def on_connection_established(self, pipe):
        self.pipe = pipe
        self.exit_stage()

    def on_exit_stage(self):
        game_stage = MultiplayerClientGameStage(
                self.world, self.gui_actor, self.pipe)
        game_stage.successor.successor = PostgameSplashStage()
        self.successor = game_stage


class ServerConnectionStage(Stage):

    def __init__(self, world, referee, num_clients, ai_actors=None,
            host=DEFAULT_HOST, port=DEFAULT_PORT):
        super().__init__()
        self.world = world
        self.referee = referee
        self.ai_actors = ai_actors or []
        self.host = host
        self.port = port
        self.pipes = []
        self.greetings = []
        self.server = linersock.Server(
                host, port, num_clients, self.on_clients_connected)

    def on_enter_stage(self):
        self.server.open()

    def on_update_stage(self, dt):
        if not self.server.finished():
            self.server.accept()
        else:
            self.exit_stage()

    def on_clients_connected(self, pipes):
        self.pipes += pipes

    def on_exit_stage(self):
        self.successor = MultiplayerServerGameStage(
                self.world, self.referee, self.ai_actors, self.pipes)


class PostgameSplashStage(Stage):
    """
    Until the player closes the window, keep it as it was when the game ended.
    """

    def on_update_stage(self, dt):
        try:
            self.gui.on_refresh_gui()
        except AttributeError:
            self.exit_stage()


class ProcessPool:
    """
    Manage a group of processes running instances of the game loop.

    This class wraps around the basic multiprocessing functionality available 
    in the python standard library.  There are two nice features provided by 
    this class.  The first is that log messages generated in the subprocesses 
    are automatically relayed to the main process, where they are reported 
    with the name of the original process included and without any mangling due 
    to race conditions.  The second is that exceptions, which are usually 
    silently ignored in subprocesses, are also relayed to the main process and 
    re-raised.

    This class is actually not specific to the game engine at all, so in 
    principle it could be moved into it's own library.  I decided not to do 
    that because I can't think of any other scenario where I would want the 
    functionality that this class offers, but maybe I'll think of one later.
    """

    def __init__(self, time_limit=None, frame_rate=30):
        self.log_queue = multiprocessing.Queue()
        self.exception_queue = multiprocessing.Queue()
        self.time_limit = time_limit
        self.elapsed_time = 0
        self.frame_rate = frame_rate

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._run_supervisor()


    def start(self, name, worker, *args, **kwargs):
        process = multiprocessing.Process(
                name=name, target=self._run_worker,
                args=(name, worker) + args, kwargs=kwargs,
        )
        process.start()

    def _run_worker(self, name, worker, *args, **kwargs):

        # Configure all logging message generated by this process to go into a 
        # queue that will be read and handled by the supervisor.

        handler = logging.handlers.QueueHandler(self.log_queue)
        logging.root.addHandler(handler)

        # Catch any exceptions generated by the worker and report them to the 
        # supervisor.  This is important, because otherwise they would be 
        # silently ignored.

        try:
            worker(*args, **kwargs)
        except Exception as exception:
            self.exception_queue.put_nowait(exception)

    def _run_supervisor(self):
        """
        Poll the queues that the worker can use to communicate with the 
        supervisor, until all the workers are done and all the queues are 
        empty.  Handle messages as they appear.
        """
        import time

        still_supervising = lambda: (
                multiprocessing.active_children()
                or not self.log_queue.empty()
                or not self.exception_queue.empty())

        try:
            while still_supervising():
                # When a log message is received, make a logger with the same 
                # name in this process and use it to re-log the message.  It 
                # will get handled in this process.

                try:
                    record = self.log_queue.get_nowait()
                    logger = logging.getLogger(record.name)
                    logger.handle(record)
                except queue.Empty:
                    pass

                # When an exception is received, immediately re-raise it.

                try:
                    exception = self.exception_queue.get_nowait()
                except queue.Empty:
                    pass
                else:
                    raise exception

                # Sleep for a little bit, and make sure that the workers haven't 
                # outlived their time limit.

                time.sleep(1/self.frame_rate)
                self.elapsed_time += 1/self.frame_rate

                if self.time_limit and self.elapsed_time > self.time_limit:
                    raise RuntimeError("timeout")

        # Make sure the workers don't outlive the supervisor, no matter how the 
        # polling loop ended (e.g. normal execution or an exception).

        finally:
            for process in multiprocessing.active_children():
                process.terminate()


class MultiplayerDebugger:
    """
    Simultaneously plays any number of different game theaters, executing each 
    theater in its own process.  This greatly facilitates the debugging and 
    testing multiplayer games.
    """

    def __init__(self, world_cls, referee_cls, gui_cls, gui_actor_cls,
            num_guis=2, ai_actor_cls=None, num_ais=0, theater_cls=PygletTheater,
            host=DEFAULT_HOST, port=DEFAULT_PORT, log_format=
            '%(levelname)s: %(processName)s: %(name)s: %(message)s'):

        # Members of this class have to be pickle-able, because this object 
        # will be pickled and sent to every process that gets started.  That's 
        # why all the game objects are stored as classes (or factories) rather 
        # than instances.  Even though some of the game objects can be pickled, 
        # none of them are meant to be and avoiding it reduces the risk that 
        # things will break for strange reasons.  The game objects themselves 
        # are instantiated in the worker processes, which is how it would 
        # happen if the user just rame multiple instances of the game anyway.

        self.theater_cls = theater_cls
        self.world_cls = world_cls
        self.referee_cls = referee_cls
        self.gui_cls = gui_cls
        self.gui_actor_cls = gui_actor_cls
        self.num_guis = num_guis
        self.ai_actor_cls = ai_actor_cls
        self.num_ais = num_ais
        self.host = host
        self.port = port
        self.log_format = log_format

    def play(self):
        # Configure the logging system to print to stderr and include the 
        # process name in all of its messages.

        handler = logging.StreamHandler()
        formatter = logging.Formatter(self.log_format)
        handler.setFormatter(formatter)
        logging.root.addHandler(handler)

        # Run the server and the client (each in its own process).

        with ProcessPool() as pool:
            pool.start("Server", self.play_server)
            for i in range(self.num_guis):
                pool.start("Client #%d" % i, self.play_client)

    def play_server(self):
        # Defer instantiation of all the game objects until we're inside our 
        # own process, to avoid having to pickle and unpickle things that 
        # should be pickled.

        theater = self.theater_cls()
        theater.initial_stage = ServerConnectionStage(
                world=self.world_cls(),
                referee=self.referee_cls(),
                num_clients=self.num_guis,
                ai_actors=[self.ai_actor_cls() for i in range(self.num_ais)],
                host=self.host,
                port=self.port,
        )
        theater.play()

    def play_client(self):
        # Defer instantiation of all the game objects until we're inside our 
        # own process, to avoid having to pickle and unpickle things that 
        # should be pickled.

        theater = self.theater_cls()
        theater.gui = self.gui_cls()
        theater.initial_stage = ClientConnectionStage(
                world=self.world_cls(),
                gui_actor=self.gui_actor_cls(),
                host=self.host,
                port=self.port,
        )
        theater.play()



def main(world_cls, referee_cls, gui_cls, gui_actor_cls, ai_actor_cls,
        theater_cls=PygletTheater, default_host=DEFAULT_HOST,
        default_port=DEFAULT_PORT, argv=None):
    """
Run a game being developed with the kxg game engine.

Usage:
    {exe_name} sandbox [<num_ais>] [-v...]
    {exe_name} client [--host HOST] [--port PORT] [-v...]
    {exe_name} server <num_guis> [<num_ais>] [options] [-v...] 
    {exe_name} debug <num_guis> [<num_ais>] [options] [-v...]
    {exe_name} --help

Commands:
    sandbox
        Play a single-player game with the specified number of AIs.  None of 
        the multiplayer machinery will be used.

    client
        Launch a client that will try to connect to a server on the given host 
        and port.  Once it connects and the game starts, the client will allow 
        you to play the game against any other connected clients.

    server
        Launch a server that will manage a game between the given number of 
        human and AI players.  The human players must connect using this 
        command's client mode.

    debug
        Debug a multiplayer game locally.  This command launches a server and 
        the given number of clients all in different processes, and configures 
        the logging system such that the output from each process can be easily 
        distinguished.

Arguments:
    <num_guis>
        The number of human players that will be playing the game.  Only needed 
        by commands that will launch some sort of multiplayer server.

    <num_ais>
        The number of AI players that will be playing the game.  Only needed by 
        commands that will launch single-player games or multiplayer servers.

Options:
    -x --host HOST          [default: {default_host}]
        The address of the machine running the server.  Must be accessible from 
        the machines running the clients.

    -p --port PORT          [default: {default_port}]
        The port that the server should listen on.  Don't specify a value less 
        than 1024 unless the server is running with root permissions.

    -v --verbose 
        Have the game engine log more information about what it's doing.  You 
        can specify this option several times to get more and more information.

This command is provided so that you can start writing your game with the least 
possible amount of boilerplate code.  However, the clients and servers provided 
by this command are not capable of running a production game.  Once you have 
written your game and want to give it a polished set of menus and options, 
you'll have to write new Stage subclasses encapsulating that logic and you'll 
have to call those stages yourself by interacting more directly with the 
Theater class.  The online documentation has more information on this process.
    """
    import sys, os, docopt, nonstdlib

    exe_name = os.path.basename(sys.argv[0])
    usage = main.__doc__.format(**locals()).strip()
    args = docopt.docopt(usage, argv or sys.argv[1:])
    num_guis = int(args['<num_guis>'] or 1)
    num_ais = int(args['<num_ais>'] or 0)
    host, port = args['--host'], int(args['--port'])

    logging.basicConfig(
            format='%(levelname)s: %(name)s: %(message)s',
            level=nonstdlib.verbosity(args['--verbose']),
    )

    # Use the given game objects and command line arguments to play a game!

    if args['debug']:
        print("""\
****************************** KNOWN BUG WARNING ******************************
In debug mode, every message produced by the logging system gets printed twice.
I know vaguely why this is happening, but as of yet I've not been able to fix
it.  In the mean time, don't let this confuse you!
*******************************************************************************""")
        game = MultiplayerDebugger(
                world_cls, referee_cls, gui_cls, gui_actor_cls, num_guis,
                ai_actor_cls, num_ais, theater_cls, host, port)
    else:
        game = theater_cls()
        ai_actors = [ai_actor_cls() for i in range(num_ais)]

        if args['sandbox']:
            game.gui = gui_cls()
            game.initial_stage = UniplayerGameStage(
                    world_cls(), referee_cls(), gui_actor_cls(), ai_actors)
            game.initial_stage.successor = PostgameSplashStage()

        if args['client']:
            game.gui = gui_cls()
            game.initial_stage = ClientConnectionStage(
                    world_cls(), gui_actor_cls(), host, port)

        if args['server']:
            game.initial_stage = ServerConnectionStage(
                    world_cls(), referee_cls(), num_guis, ai_actors,
                    host, port)

    game.play()




#!/usr/bin/env python3

from .errors import *
from .forums import ForumObserver
from .actors import require_actor

def read_only(method):
    setattr(method, '_kxg_read_only', True)
    return method

def before_world(method):
    setattr(method, '_kxg_before_world', True)
    return method

def watch_token(method):
    """
    Mark a token extension method that should automatically be called when a 
    token method of the same name is called.

    This decorator must only be used on TokenExtension methods, otherwise it 
    will silently do nothing.  The reason is that the decorator itself can't do 
    anything but label the given method, because at the time of decoration the 
    token to watch isn't known.  The method is actually setup to watch a token 
    in the TokenExtension constructor, which searches for the label added here.  
    But other classes won't make this search and will silently do nothing.
    """
    method._kxg_watch_token = True
    return method

@debug_only
def require_token(object):
    return require_instance(Token(), object)

@debug_only
def require_active_token(object):
    token = require_token(object)

    if token.world_registration == 'pending':
        if not token.has_id():
            raise TokenDoesntHaveId(token)
        if not token.has_world():
            raise TokenNotInWorld(token)
    if token.world_registration == 'expired':
        raise UsingRemovedToken(token)

    return token


class TokenMetaclass(type):

    def __new__(mcs, name, bases, members):
        """
        Add checks to make sure token methods are being called safely.

        In order to keep multiplayer games in sync, the world should only be 
        modified at particular times (e.g. token update methods and messages).  
        The purpose of this metaclass is to stop you from accidentally trying 
        to modify the world outside of these defined times.  These mistakes 
        would otherwise cause hard-to-debug sync errors.

        The engine indicates when it is safe to modify the world by setting a 
        boolean lock flag in the world.  This metaclass adds a bit of logic to 
        non-read-only token methods that makes sure the world is unlocked 
        before continuing.  The kxg.read_only() decorator can be used to 
        indicate which methods are read-only, and are therefore excluded from 
        these checks.
        
        The checks configured by this metaclass help find bugs, but may also 
        incur significant computational expense.  By invoking python with 
        optimization enabled (i.e. passing -O) these checks are skipped.
        """

        if __debug__:
            mcs.add_safety_checks(members)

        return super().__new__(mcs, name, bases, members)
        

    @classmethod
    def add_safety_checks(mcs, members):
        """
        Iterate through each member of the class being created and add a 
        safety check to every method that isn't marked as read-only.
        """
        for member_name, member_value in members.items():
            members[member_name] = mcs.add_safety_check(
                    member_name, member_value)

    @staticmethod
    def add_safety_check(member_name, member_value):
        """
        If the given member is a method that hasn't been marked as read-only, 
        return a version of it that will complain if invoked in a dangerous 
        way.  This mostly means checking to make sure that methods that alter 
        the token are only called from update methods or messages.
        """

        import functools
        from types import FunctionType

        is_method = isinstance(member_value, FunctionType)
        is_read_only = hasattr(member_value, '_kxg_read_only')
        is_engine_helper = member_name.startswith('_')

        if not is_method or is_read_only or is_engine_helper:
            return member_value

        def safety_checked_method(self, *args, **kwargs):

            if self.world_registration == 'pending':

                # Calling a non-read-only method before on a token before 
                # adding it to the world isn't inherently a synchronization 
                # issue, because until a token is added to the world only 
                # one client knows about it.  However, this happens most 
                # often when the user forgets to add a token to the world 
                # in the first place.  To catch these bugs, non-read-only 
                # token methods can't be called before the token has been 
                # added to the world unless they are explicitly labeled 
                # with the kxg.before_world() decorator.

                if not hasattr(member_value, '_kxg_before_world'):
                    raise CantModifyTokenIfNotInWorld(self, member_name)

            if self.world_registration == 'active':

                # If the token has already been added to the world, make 
                # sure that the token seems properly set up (specifically 
                # that it has all the attributes that tokens should have 
                # and that it's id isn't null) and that the world is 
                # unlocked.

                require_active_token(self)

                if self.world.is_locked():
                    raise CantModifyTokenIfWorldLocked(self, member_name)

            if self.world_registration == 'expired':

                # Once a token has been removed from the world, almost 
                # anything you do with it will raise an exception.  This 
                # behavior is meant to prevent bugs that happen when you 
                # accidentally keep stale references to removed tokens.  If 
                # you need to add a removed token again, you can call its 
                # reset_registration() method, which will allow you to add 
                # it to the world again in the usual way (i.e. by sending a 
                # message).

                if member_name != 'reset_registration':
                    raise UsingRemovedToken(self)

            # After all the checks have been carried out, call the method 
            # as usual.

            return member_value(self, *args, **kwargs)

        # Preserve any "forum observer" decorations that have been placed on 
        # the method.  Also restore the method's original name and module 
        # strings, to make inspection and debugging a little easier.

        functools.update_wrapper(
                safety_checked_method, member_value,
                assigned=functools.WRAPPER_ASSIGNMENTS + (
                    '_kxg_subscribe_to_message',
                    '_kxg_subscribe_to_sync_response',
                    '_kxg_subscribe_to_undo_response',
                )
        )
        return safety_checked_method

        

class Token(ForumObserver, metaclass=TokenMetaclass):

    class WatchedMethod:

        def __init__(self, method):
            self.method = method
            self.watchers = []

        def __call__(self, *args, **kwargs):
            self.method(*args, **kwargs)
            for watcher in self.watchers:
                watcher(*args, **kwargs)

        def add_watcher(self, watcher):
            self.watchers.append(watcher)


    def __init__(self):
        super().__init__()
        self._id = None
        self._world = None
        self._removed_from_world = False
        self._extensions = {}
        self._disable_forum_observation()

    def __repr__(self):
        return '{}(id={})'.format(
                self.__class__.__name__, getattr(self, 'id', None))

    def __getstate__(self):
        state = super().__getstate__()
        del state['_world']
        del state['_removed_from_world']
        del state['_extensions']
        return state

    def __setstate__(self, state):
        Token.__init__(self)
        super().__setstate__(state)

    def __extend__(self):
        return {}

    @property
    @read_only
    def id(self):
        return self._id

    @property
    @read_only
    def world(self):
        return self._world

    @property
    @read_only
    def world_registration(self):
        """
        Return the status of this token's registration with the world.

        There are three possible world registration statuses.  The first is 
        'pending', which means that the token has not yet been added to the 
        world (although it may have been assigned an id).  The second is 
        'active', which means that the token is fully participating in the 
        game.  The third is 'expired', which means that the token has been 
        removed from the world and should no longer be in use.  These statuses 
        are mostly used for internal checks within the game engine.
        """
        if self._removed_from_world:
            return 'expired'
        elif self.has_world():
            if not self.has_id():
                raise TokenDoesntHaveId(self)
            return 'active'
        else:
            return 'pending'

    @read_only
    def has_id(self):
        return self.id is not None

    @read_only
    def has_world(self):
        assert (not self.world) or (self in self.world)
        return self.world is not None

    @read_only
    def has_extension(self, actor):
        require_actor(actor)
        return actor in self._extensions

    @read_only
    def get_extension(self, actor):
        require_actor(actor)
        return self._extensions[actor]

    @read_only
    def get_extensions(self):
        return list(self._extensions.values())

    @read_only
    def watch_method(self, method_name, callback):
        """
        Register the given callback to be called whenever the method with the 
        given name is called.  You can easily take advantage of this feature in 
        token extensions by using the @watch_token decorator.
        """

        # Make sure a token method with the given name exists, and complain if 
        # nothing is found.

        try:
            method = getattr(self, method_name)
        except AttributeError:
            raise TokenHasNoSuchMethodToWatch(self, method_name)

        # Wrap the method in a WatchedMethod object, if that hasn't already 
        # been done.  This object manages a list of callback method and takes 
        # responsibility for calling them after the method itself has been 
        # called.

        if not isinstance(method, Token.WatchedMethod):
            setattr(self, method_name, Token.WatchedMethod(method))
            method = getattr(self, method_name)

        # Add the given callback to the watched method.

        method.add_watcher(callback)

    @read_only
    def reset_registration(self):
        """
        Allow the token to be added to the world again.

        Once a token has been removed from the world, almost anything you do 
        with it will raise an exception.  This behavior is meant to prevent 
        bugs that happen when you accidentally keep stale references to removed 
        tokens.  Calling this method circumvents those checks behavior and 
        allows you to add a removed token back into the world.  Note that this 
        method does not actually add the token to the world, it just allows you 
        to do that in the usual way (i.e. by sending a message).  
        """
        if self.world_registration != 'expired':
            raise CantResetActiveToken(self)
        Token.__init__(self)

    def on_add_to_world(self, world):
        pass

    def on_update_game(self, dt):
        pass

    @read_only
    def on_report_to_referee(self, reporter):
        pass

    def on_remove_from_world(self):
        pass

    def _give_id(self, id_factory):
        from .forums import IdFactory

        require_token(self)

        if self.has_id():
            raise TokenAlreadyHasId(self)
        if not isinstance(id_factory, IdFactory):
            raise NotUsingIdFactory(id_factory)

        self._id = id_factory.next()

    def _check_if_forum_observation_enabled(self):
        """
        Give a helpful error if the user attempts to subscribe or unsubscribe 
        from messages while the token is not registered with a world.  This can 
        easily happen if the user attempts to subscribe to messages in the 
        constructor.  However, because the constructor is only called on one 
        client and message handlers cannot be pickled, subscribing at this time 
        would create hard-to-find synchronization bugs.
        """
        try:
            super()._check_if_forum_observation_enabled()
        except AssertionError:
            raise TokenCantSubscribeNow(self)


class TokenExtension(ForumObserver):

    def __init__(self, actor, token):
        super().__init__()
        self.actor = actor
        self.token = token

        # Iterate through all of the extension methods to find ones wanting to 
        # "watch" the token, then configure the token to call these methods 
        # whenever a token method of the same name is called.
        
        from inspect import getmembers, ismethod

        for method_name, method in getmembers(self, ismethod):

            # Methods with the '_kxg_watch_token' attribute set should be set 
            # up to watch the token.  This attribute is typically set using the
            # @watch_token decorator.

            if hasattr(method, '_kxg_watch_token'):
                token.watch_method(method_name, method)

    def __rshift__(self, message):
        return self.send_message(message)

    def send_message(self, message):
        return self.actor.send_message(message)


class TokenSerializer:

    def __init__(self, world):
        self.world = world

    def pack(self, message):
        from pickle import Pickler
        from io import BytesIO

        buffer = BytesIO()
        delegate = Pickler(buffer)

        def persistent_id(token):
            if isinstance(token, Token):
                if token.world_registration == 'pending':
                    return None
                if token.world_registration == 'active':
                    require_active_token(token)
                    return token.id
                if token.world_registration == 'expired':
                    raise UsingRemovedToken(token)

        delegate.persistent_id = persistent_id
        delegate.dump(message)

        return buffer.getvalue()

    def unpack(self, packet):
        from pickle import Unpickler
        from io import BytesIO

        buffer = BytesIO(packet)
        delegate = Unpickler(buffer)

        delegate.persistent_load = lambda id: self.world.get_token(int(id))
        return delegate.load()






from . import quickstart
from .theater import *
from .forums import *
from .actors import *
from .multiplayer import *
from .messages import *
from .tokens import *
from .world import *
from .errors import *



import distutils.core

# Uploading to PyPI
# =================
# $ python setup.py register -r pypi
# $ python setup.py sdist upload -r pypi

version = '0.1.1'
distutils.core.setup(
        name='kxg',
        version=version,
        author='Kale Kundert and Alex Mitchell',
        url='https://github.com/kxgames/GameEngine',
        download_url='https://github.com/kxgames/GameEngine/tarball/'+version,
        license='LICENSE.txt',
        description="A multiplayer game engine.",
        long_description=open('README.rst').read(),
        keywords=['game', 'network', 'gui', 'pyglet'],
        packages=['kxg'],
        install_requires=[
            'pyglet',
            'nonstdlib',
            'linersock',
            'pytest',
            'docopt',
        ],
)
